"use strict";
var index_1 = require("./index");
var TransformOperationExecutor = (function () {
    // -------------------------------------------------------------------------
    // Constructor
    // -------------------------------------------------------------------------
    function TransformOperationExecutor(transformationType, options) {
        this.transformationType = transformationType;
        this.options = options;
        // -------------------------------------------------------------------------
        // Private Properties
        // -------------------------------------------------------------------------
        this.transformedTypes = [];
    }
    // -------------------------------------------------------------------------
    // Public Methods
    // -------------------------------------------------------------------------
    TransformOperationExecutor.prototype.transform = function (source, value, targetType, arrayType, level) {
        var _this = this;
        if (level === void 0) { level = 0; }
        if (value instanceof Array) {
            var newValue_1 = arrayType && this.transformationType === "plainToClass" ? new arrayType() : [];
            value.forEach(function (subValue, index) {
                var subSource = source ? source[index] : undefined;
                if (!_this.isCircular(subValue, level)) {
                    newValue_1.push(_this.transform(subSource, subValue, targetType, undefined, level + 1));
                }
                else if (_this.transformationType === "classToClass") {
                    newValue_1.push(subValue);
                }
            });
            return newValue_1;
        }
        else if (targetType === String) {
            return String(value);
        }
        else if (targetType === Number) {
            return Number(value);
        }
        else if (targetType === Boolean) {
            return Boolean(value);
        }
        else if (targetType === Date || value instanceof Date) {
            if (value instanceof Date) {
                return new Date(value.getTime());
            }
            return new Date(value);
        }
        else if (value instanceof Object) {
            // try to guess the type
            if (!targetType && value.constructor !== Object /* && operationType === "classToPlain"*/)
                targetType = value.constructor;
            if (!targetType && source)
                targetType = source.constructor;
            // add transformed type to prevent circular references
            this.transformedTypes.push({ level: level, object: value });
            var keys = this.getKeys(targetType, value);
            var newValue = source ? source : {};
            if (!source && (this.transformationType === "plainToClass" || this.transformationType === "classToClass")) {
                newValue = new targetType();
            }
            // traverse over keys
            for (var _i = 0, keys_1 = keys; _i < keys_1.length; _i++) {
                var key = keys_1[_i];
                var valueKey = key, newValueKey = key, propertyName = key;
                if (!this.options.ignoreDecorators) {
                    if (this.transformationType === "plainToClass") {
                        var exposeMetadata = index_1.defaultMetadataStorage.findExposeMetadataByCustomName(targetType, key);
                        if (exposeMetadata) {
                            propertyName = exposeMetadata.propertyName;
                            newValueKey = exposeMetadata.propertyName;
                        }
                    }
                    else if (this.transformationType === "classToPlain" || this.transformationType === "classToClass") {
                        var exposeMetadata = index_1.defaultMetadataStorage.findExposeMetadata(targetType, key);
                        if (exposeMetadata && exposeMetadata.options && exposeMetadata.options.name)
                            newValueKey = exposeMetadata.options.name;
                    }
                }
                var type = this.getKeyType(value, targetType, propertyName);
                // if value is an array try to get its custom array type
                var arrayType_1 = value[valueKey] instanceof Array ? this.getReflectedType(targetType, propertyName) : undefined;
                // const subValueKey = operationType === "plainToClass" && newKeyName ? newKeyName : key;
                var subValue = value[valueKey] instanceof Function ? value[valueKey]() : value[valueKey];
                var subSource = source ? source[valueKey] : undefined;
                // if its deserialization then type if required
                if (this.transformationType === "plainToClass" && !type && subValue instanceof Object && !(subValue instanceof Date))
                    throw new Error("Cannot determine type for " + targetType.name + "." + propertyName + ", did you forget to specify a @Type?");
                // if newValue is a source object that has method that match newKeyName then skip it
                var hasDescriptor = false;
                if (newValue.constructor.prototype) {
                    var descriptor = Object.getOwnPropertyDescriptor(newValue.constructor.prototype, newValueKey);
                    hasDescriptor = !!descriptor;
                    if ((this.transformationType === "plainToClass" || this.transformationType === "classToClass") && (newValue[newValueKey] instanceof Function || descriptor))
                        continue;
                }
                if (!this.isCircular(subValue, level)) {
                    newValue[newValueKey] = this.transform(subSource, subValue, type, arrayType_1, level + 1);
                }
                else if (this.transformationType === "classToClass") {
                    newValue[newValueKey] = subValue;
                }
            }
            return newValue;
        }
        else {
            return value;
        }
    };
    // preventing circular references
    TransformOperationExecutor.prototype.isCircular = function (object, level) {
        return !!this.transformedTypes.find(function (transformed) { return transformed.object === object && transformed.level < level; });
    };
    TransformOperationExecutor.prototype.getKeyType = function (object, target, key) {
        if (!target)
            return undefined;
        var metadata = index_1.defaultMetadataStorage.findTypeMetadata(target, key);
        return metadata ? metadata.typeFunction(object) : undefined;
    };
    TransformOperationExecutor.prototype.getReflectedType = function (target, propertyName) {
        if (!target)
            return undefined;
        var meta = index_1.defaultMetadataStorage.findTypeMetadata(target, propertyName);
        return meta ? meta.reflectedType : undefined;
    };
    TransformOperationExecutor.prototype.getKeys = function (target, object) {
        var _this = this;
        // determine exclusion strategy
        var strategy = index_1.defaultMetadataStorage.getStrategy(target);
        if (strategy === "none")
            strategy = this.options.strategy || "exposeAll"; // exposeAll is default strategy
        // get all keys that need to expose
        var keys = strategy === "exposeAll" ? Object.keys(object) : [];
        if (!this.options.ignoreDecorators) {
            // add all exposed to list of keys
            var exposedProperties = index_1.defaultMetadataStorage.getExposedProperties(target, this.transformationType);
            if (this.transformationType === "plainToClass") {
                exposedProperties = exposedProperties.map(function (key) {
                    var exposeMetadata = index_1.defaultMetadataStorage.findExposeMetadata(target, key);
                    if (exposeMetadata && exposeMetadata.options && exposeMetadata.options.name) {
                        return exposeMetadata.options.name;
                    }
                    return key;
                });
            }
            keys = keys.concat(exposedProperties);
            // exclude excluded properties
            var excludedProperties_1 = index_1.defaultMetadataStorage.getExcludedProperties(target, this.transformationType);
            if (excludedProperties_1.length > 0) {
                keys = keys.filter(function (key) {
                    return excludedProperties_1.indexOf(key) === -1;
                });
            }
            // apply versioning options
            if (this.options.version !== undefined) {
                keys = keys.filter(function (key) {
                    var exposeMetadata = index_1.defaultMetadataStorage.findExposeMetadata(target, key);
                    if (!exposeMetadata || !exposeMetadata.options)
                        return true;
                    var decision = true;
                    if (decision && exposeMetadata.options.since)
                        decision = _this.options.version >= exposeMetadata.options.since;
                    if (decision && exposeMetadata.options.until)
                        decision = _this.options.version < exposeMetadata.options.until;
                    return decision;
                });
            }
            // apply grouping options
            if (this.options.groups && this.options.groups.length) {
                keys = keys.filter(function (key) {
                    var exposeMetadata = index_1.defaultMetadataStorage.findExposeMetadata(target, key);
                    if (!exposeMetadata || !exposeMetadata.options || !exposeMetadata.options.groups)
                        return true;
                    return _this.options.groups.some(function (optionGroup) { return exposeMetadata.options.groups.indexOf(optionGroup) !== -1; });
                });
            }
            else {
                keys = keys.filter(function (key) {
                    var exposeMetadata = index_1.defaultMetadataStorage.findExposeMetadata(target, key);
                    return !exposeMetadata ||
                        !exposeMetadata.options ||
                        !exposeMetadata.options.groups ||
                        !exposeMetadata.options.groups.length;
                });
            }
        }
        // exclude prefixed properties
        if (this.options.excludePrefixes && this.options.excludePrefixes.length) {
            keys = keys.filter(function (key) { return _this.options.excludePrefixes.every(function (prefix) {
                return key.substr(0, prefix.length) !== prefix;
            }); });
        }
        // make sure we have unique keys
        keys = keys.filter(function (key, index, self) {
            return self.indexOf(key) === index;
        });
        return keys;
    };
    return TransformOperationExecutor;
}());
exports.TransformOperationExecutor = TransformOperationExecutor;

//# sourceMappingURL=TransformOperationExecutor.js.map
